# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Puppeteers for gift_refinements."""

from collections.abc import Mapping

import dm_env
import numpy as np
import tree

from meltingpot.utils.puppeteers import puppeteer

Observation = Mapping[str, tree.Structure[np.ndarray]]


class GiftRefinementsCooperator(puppeteer.Puppeteer[tuple[()]]):
  """Cooperator puppeteer for gift refinements.

  This puppeteer expresses a cooperative high level policy:
  1.  Collect tokens when the inventory is empty.
  2.  If the inventory is not empty, check if there are any refined tokens, if
      not, the gift some tokens.
  3.  If there the player has refined tokens, consume.

  This means that a GiftRefinementsCooperator will start by grabbing a token,
  and then gift it. As soon as they receive any gift from anyone, they would
  consume.
  """

  def __init__(
      self,
      *,
      collect_goal: puppeteer.PuppetGoal,
      gift_goal: puppeteer.PuppetGoal,
      consume_goal: puppeteer.PuppetGoal,
  ):
    """Initializes the puppeteer.

    Args:
      collect_goal: goal to emit to puppet when "collecting"
      gift_goal: goal to emit to puppet when "gifting"
      consume_goal: goal to emit to puppet when "consuming"
    """
    self._collect_goal = collect_goal
    self._gift_goal = gift_goal
    self._consume_goal = consume_goal

  def initial_state(self) -> tuple[()]:
    """See base class."""
    return ()

  def should_consume(self, observation: Observation) -> bool:
    """Decides whether we should consume tokens in our inventory."""
    _, refined, twice_refined = observation['INVENTORY']
    return bool(refined) or bool(twice_refined)

  def step(self, timestep: dm_env.TimeStep,
           prev_state: tuple[()]) -> tuple[dm_env.TimeStep, tuple[()]]:
    """See base class."""
    if np.sum(timestep.observation['INVENTORY']):
      if self.should_consume(timestep.observation):
        goal = self._consume_goal
      else:
        goal = self._gift_goal
    else:
      goal = self._collect_goal

    # Return the encoded cumulant associated with the current goal.
    timestep = puppeteer.puppet_timestep(timestep, goal)
    return timestep, prev_state


class GiftRefinementsExtremeCooperator(GiftRefinementsCooperator):
  """Cooperator that gifts until it has tokens of type 2 (double refinement).

  This means that a GiftRefinementsExtremeCooperator, like the cooperator above,
  will start by grabbing a token, and then gift it. However, upon receiving a
  gift, they would gift back. Only will they consume if they receive a doubly
  refined token.
  """

  def should_consume(self, observation: Observation) -> bool:
    """See base class."""
    _, _, twice_refined = observation['INVENTORY']
    return bool(twice_refined > 0)



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Collaborative Cooking: Figure Eight.

Example video: https://youtu.be/hUCbOL5l-Gw

The recipe they must follow is for tomato soup:
1.   Add three tomatoes to the cooking pot.
2.   Wait for the soup to cook (status bar completion).
3.   Bring a bowl to the pot and pour the soup from the pot into the bowl.
4.   Deliver the bowl of soup at the goal location.

This substrate is a pure common interest game. All players share all rewards.

Players have a `5 x 5` observation window.

Map:
Figure Eight: The map is a figure eight shaped maze that generates numerous
places where players might get in one another's way, blocking critical paths.
While it is technically possible for a single player to complete the task on
their own it is very unlikely that poor performing partners would get out of its
way, so in practice, collaboration is essential.
"""

from ml_collections import config_dict

from meltingpot.configs.substrates import collaborative_cooking as base_config
from meltingpot.utils.substrates import specs

build = base_config.build

# Figure Eight: Strong performance on this map requires two stages of teamwork.
ASCII_MAP = """
################
####C#C##C#C####
# P          P #
## ########## ##
#    P   P     #
## ########## ##
#    P   P     #
### #ODTTOD# ###
################
"""


def get_config():
  """Default configuration."""
  config = base_config.get_config()

  # Override the map layout settings.
  config.layout = config_dict.ConfigDict()
  config.layout.ascii_map = ASCII_MAP
  # The specs of the environment (from a single-agent perspective).
  config.timestep_spec = specs.timestep({
      "RGB": specs.rgb(40, 40),
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(72, 128),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"default"})
  config.default_player_roles = ("default",) * 6

  return config



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for Collaborative Cooking: Cramped.

Example video: https://youtu.be/8qQFbxO8UNY

The recipe they must follow is for tomato soup:
1.   Add three tomatoes to the cooking pot.
2.   Wait for the soup to cook (status bar completion).
3.   Bring a bowl to the pot and pour the soup from the pot into the bowl.
4.   Deliver the bowl of soup at the goal location.

This substrate is a pure common interest game. All players share all rewards.

Players have a `5 x 5` observation window.

Map:
Cramped Room: A tight layout requiring significant movement coordination between
the players in order to avoid being blocked by each other.
"""

from ml_collections import config_dict

from meltingpot.configs.substrates import collaborative_cooking as base_config
from meltingpot.utils.substrates import specs

build = base_config.build

# Cramped Room: A tight layout requiring significant movement coordination
# between the players in order to avoid being blocked by each other.
ASCII_MAP = """
xx##C##xx
xxOP POxx
xx#   #xx
xx#D#T#xx
xxxxxxxxx
"""


def get_config():
  """Default configuration."""
  config = base_config.get_config()

  # Override the map layout settings.
  config.layout = config_dict.ConfigDict()
  config.layout.ascii_map = ASCII_MAP
  # The specs of the environment (from a single-agent perspective).
  config.timestep_spec = specs.timestep({
      "RGB": specs.rgb(40, 40),
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(40, 72),
  })

  # The roles assigned to each player.
  config.valid_roles = frozenset({"default"})
  config.default_player_roles = ("default",) * 2

  return config



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for in_the_matrix Puppeteers."""

import itertools
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import dm_env
import immutabledict
import numpy as np

from meltingpot.utils.puppeteers import in_the_matrix
from meltingpot.utils.puppeteers import testutils

_RESOURCE_0 = in_the_matrix.Resource(
    index=0,
    collect_goal=mock.sentinel.collect_0,
    interact_goal=mock.sentinel.interact_0,
)
_RESOURCE_1 = in_the_matrix.Resource(
    index=1,
    collect_goal=mock.sentinel.collect_1,
    interact_goal=mock.sentinel.interact_1,
)
_RESOURCE_2 = in_the_matrix.Resource(
    index=2,
    collect_goal=mock.sentinel.collect_2,
    interact_goal=mock.sentinel.interact_2,
)

_INTERACTION = (np.array([1, 1, 1]), np.array([1, 1, 1]))
_NO_INTERACTION = (np.array([-1, -1, -1]), np.array([-1, -1, -1]))


class HelperFunctionTest(parameterized.TestCase):

  @parameterized.parameters(
      [_INTERACTION, _INTERACTION[1]],
      [_NO_INTERACTION, None],
  )
  def test_get_partner_interaction_inventory(self, interaction, expected):
    timestep = dm_env.restart(_observation(None, interaction))
    actual = in_the_matrix.get_partner_interaction_inventory(timestep)
    np.testing.assert_equal(actual, expected)

  @parameterized.parameters(
      [_INTERACTION, True],
      [_NO_INTERACTION, False],
  )
  def test_has_interaction(self, interaction, expected):
    timestep = dm_env.restart(_observation(None, interaction))
    actual = in_the_matrix.has_interaction(timestep)
    self.assertEqual(actual, expected)

  @parameterized.parameters(
      [(0, 0, 0), (mock.ANY, 0)],
      [(1, 0, 0), (0, 1)],
      [(0, 3, 1), (1, 2)],
      [(3, 0, 7), (2, 4)],
  )
  def test_max_resource_and_margin(self, inventory, expected):
    actual = in_the_matrix.max_resource_and_margin(np.array(inventory))
    self.assertEqual(actual, expected)

  @parameterized.parameters(
      [(1, 2, 3), 0, 1, False],
      [(1, 2, 3), 1, 1, False],
      [(1, 2, 3), 2, 1, True],
      [(1, 2, 3), 2, 2, False],
      [(1, 2, 5), 2, 1, True],
      [(1, 2, 5), 2, 3, True],
  )
  def test_has_sufficient(self, inventory, target, margin, expected):
    actual = in_the_matrix.has_collected_sufficient(
        np.array(inventory), target, margin)
    self.assertEqual(actual, expected)

  @parameterized.parameters(
      [(1, 2, 3), 2],
      [(1, 2, 5), 2],
      [(1, 0, 0), 0],
      [(1, 2, 0), 1],
  )
  def test_partner_max_resource(self, inventory, expected):
    timestep = dm_env.restart({
        'INTERACTION_INVENTORIES': (None, np.array(inventory)),
    })
    actual = in_the_matrix.partner_max_resource(timestep)
    self.assertEqual(actual, expected)

  @parameterized.parameters(
      [(1, 2, 3), _RESOURCE_0, 1, _RESOURCE_0.collect_goal],
      [(1, 2, 3), _RESOURCE_1, 1, _RESOURCE_1.collect_goal],
      [(1, 2, 3), _RESOURCE_2, 1, _RESOURCE_2.interact_goal],
      [(1, 2, 3), _RESOURCE_2, 2, _RESOURCE_2.collect_goal],
      [(1, 2, 5), _RESOURCE_2, 1, _RESOURCE_2.interact_goal],
      [(1, 2, 5), _RESOURCE_2, 3, _RESOURCE_2.interact_goal],
  )
  @mock.patch.object(immutabledict, 'immutabledict', dict)
  def test_collect_or_interact_puppet_timestep(
      self, inventory, target, margin, goal):
    timestep = dm_env.restart({'INVENTORY': np.array(inventory)})
    actual = in_the_matrix.collect_or_interact_puppet_timestep(
        timestep, target, margin)
    expected = dm_env.restart({'INVENTORY': np.array(inventory), 'GOAL': goal})
    np.testing.assert_equal(actual, expected)


def _observation(inventory, interaction=None):
  if interaction is None:
    interaction = _NO_INTERACTION
  return {
      'INVENTORY': np.array(inventory),
      'INTERACTION_INVENTORIES': np.array(interaction),
  }


def _goals_from_observations(puppeteer,
                             inventories,
                             interactions=(),
                             state=None):
  observations = []
  for inventory, interaction in itertools.zip_longest(inventories,
                                                      interactions):
    observations.append(_observation(inventory, interaction))
  return testutils.goals_from_observations(puppeteer, observations, state)


class SpecialistTest(parameterized.TestCase):

  def test(self):
    puppeteer = in_the_matrix.Specialist(
        target=_RESOURCE_1,
        margin=1,
    )
    inventories = [(1, 1, 1), (1, 2, 1), (1, 2, 2), (1, 3, 2)]
    expected = [
        _RESOURCE_1.collect_goal,
        _RESOURCE_1.interact_goal,
        _RESOURCE_1.collect_goal,
        _RESOURCE_1.interact_goal,
    ]
    actual, _ = _goals_from_observations(puppeteer, inventories)
    self.assertEqual(actual, expected)


class ScheduledFlipTest(parameterized.TestCase):

  def test(self):
    puppeteer = in_the_matrix.ScheduledFlip(
        threshold=1,
        initial_target=_RESOURCE_1,
        final_target=_RESOURCE_2,
        initial_margin=1,
        final_margin=2,
    )
    inventories = [(1, 1, 1), (1, 2, 1), (1, 2, 2), (1, 2, 4)]
    interactions = [
        _NO_INTERACTION, _NO_INTERACTION, _INTERACTION, _INTERACTION
    ]
    expected = [
        _RESOURCE_1.collect_goal,
        _RESOURCE_1.interact_goal,
        _RESOURCE_2.collect_goal,
        _RESOURCE_2.interact_goal,
    ]
    actual, _ = _goals_from_observations(puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)


class GrimTriggerTest(parameterized.TestCase):

  def test_trigger(self):
    puppeteer = in_the_matrix.GrimTrigger(
        threshold=2,
        cooperate_resource=_RESOURCE_1,
        defect_resource=_RESOURCE_0,
        margin=1,
    )
    inventories = [
        (1, 1),
        (1, 2),
        (2, 2),
        (2, 2),
        (3, 2),
        (3, 2),
    ]
    interactions = [
        ((-1, -1), (-1, -1)),  # neither
        ((-1, -1), (1, 0)),  # defection
        ((-1, -1), (0, 1)),  # cooperation
        ((-1, -1), (1, 0)),  # defection
        ((-1, -1), (0, 1)),  # cooperation
        ((-1, -1), (0, 1)),  # cooperation
    ]
    expected = [
        _RESOURCE_1.collect_goal,  # cooperate
        _RESOURCE_1.interact_goal,  # cooperate
        _RESOURCE_1.collect_goal,  # cooperate
        _RESOURCE_0.collect_goal,  # defect
        _RESOURCE_0.interact_goal,  # defect
        _RESOURCE_0.interact_goal,  # defect
    ]
    actual, _ = _goals_from_observations(puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)

  def test_not_grim_after_reset(self):
    puppeteer = in_the_matrix.GrimTrigger(
        threshold=2,
        cooperate_resource=_RESOURCE_1,
        defect_resource=_RESOURCE_0,
        margin=1,
    )
    inventories = [
        (1, 1),
        (1, 2),
        (2, 2),
        (2, 2),
        (3, 2),
        (3, 2),
    ]
    interactions = [
        ((-1, -1), (-1, -1)),  # neither
        ((-1, -1), (1, 0)),  # defection
        ((-1, -1), (0, 1)),  # cooperation
        ((-1, -1), (1, 0)),  # defection
        ((-1, -1), (0, 1)),  # cooperation
        ((-1, -1), (0, 1)),  # cooperation
    ]
    expected = [
        _RESOURCE_1.collect_goal,  # cooperate
        _RESOURCE_1.interact_goal,  # cooperate
        _RESOURCE_1.collect_goal,  # cooperate
        _RESOURCE_0.collect_goal,  # defect
        _RESOURCE_0.interact_goal,  # defect
        _RESOURCE_0.interact_goal,  # defect
    ]
    _, state = _goals_from_observations(puppeteer, inventories, interactions)
    actual, _ = _goals_from_observations(puppeteer, inventories, interactions,
                                         state)
    self.assertEqual(actual, expected)


class TitForTatTest(parameterized.TestCase):

  def test(self):
    puppeteer = in_the_matrix.TitForTat(
        cooperate_resource=_RESOURCE_1,
        defect_resource=_RESOURCE_0,
        margin=1,
        tremble_probability=0,
    )
    inventories = [
        (1, 1, 0),  # not ready to interact
        (1, 2, 0),  # ready to interact if cooperating
        (3, 2, 0),  # ready to interact if defecting
        (2, 3, 0),  # ready to interact if cooperating
        (3, 2, 0),  # ready to interact if defecting
        (2, 2, 0),  # not ready to interact
    ]
    interactions = [
        ((-1, -1, -1), (2, 2, 0)),  # coplayer cooperates and defects
        ((-1, -1, -1), (1, 0, 0)),  # coplayer defects
        ((-1, -1, -1), (0, 0, 1)),  # coplayer plays other
        ((-1, -1, -1), (0, 1, 0)),  # coplayer cooperates
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (2, 1, 1)),  # coplayer defects
    ]
    expected = [
        _RESOURCE_1.collect_goal,  # cooperate
        _RESOURCE_0.collect_goal,  # defect
        _RESOURCE_0.interact_goal,  # continue defecting
        _RESOURCE_1.interact_goal,  # cooperate
        _RESOURCE_1.collect_goal,  # continue cooperating
        _RESOURCE_0.collect_goal,  # defect
    ]
    actual, _ = _goals_from_observations(puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)

  def test_with_tremble(self):
    puppeteer = in_the_matrix.TitForTat(
        cooperate_resource=_RESOURCE_1,
        defect_resource=_RESOURCE_0,
        margin=1,
        tremble_probability=1,
    )
    inventories = [
        (1, 1, 0),  # not ready to interact
        (1, 2, 0),  # ready to interact if cooperating
        (3, 2, 0),  # ready to interact if defecting
        (2, 3, 0),  # ready to interact if cooperating
        (3, 2, 0),  # ready to interact if defecting
        (2, 2, 0),  # not ready to interact
    ]
    interactions = [
        ((-1, -1, -1), (2, 2, 0)),  # coplayer cooperates and defects
        ((-1, -1, -1), (1, 0, 0)),  # coplayer defects
        ((-1, -1, -1), (0, 0, 1)),  # coplayer plays other
        ((-1, -1, -1), (0, 1, 0)),  # coplayer cooperates
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (2, 1, 1)),  # coplayer defects
    ]
    expected = [
        _RESOURCE_0.collect_goal,  # cooperate but tremble and defect
        _RESOURCE_1.interact_goal,  # defect but tremble and cooperate
        _RESOURCE_1.collect_goal,  # continue cooperating
        _RESOURCE_0.collect_goal,  # cooperate but tremble and defect
        _RESOURCE_0.interact_goal,  # continue defecting
        _RESOURCE_1.collect_goal,  # defect but tremble and cooperate
    ]
    actual, _ = _goals_from_observations(puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)


class CorrigibleTest(parameterized.TestCase):

  def test(self):
    puppeteer = in_the_matrix.Corrigible(
        threshold=1,
        cooperate_resource=_RESOURCE_1,
        defect_resource=_RESOURCE_0,
        margin=2,
        tremble_probability=0,
    )
    inventories = [
        (2, 0, 1),  # not ready to interact
        (4, 1, 0),  # ready to interact if defecting
        (2, 1, 0),  # not ready to interact
        (1, 2, 0),  # not ready to interact
        (2, 1, 0),  # not ready to interact
        (1, 4, 0),  # ready to interact if cooperating
        (3, 1, 0),  # ready to interact if defecting
    ]
    interactions = [
        ((-1, -1, -1), (0, 1, 0)),  # coplayer cooperates
        ((-1, -1, -1), (0, 0, 1)),  # coplayer plays other
        ((-1, -1, -1), (1, 0, 0)),  # coplayer defects
        ((-1, -1, -1), (0, 1, 0)),  # coplayer cooperates
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (1, 1, 0)),  # coplayer cooperates and defects
        ((-1, -1, -1), (2, 0, 1)),  # coplayer defects
    ]
    expected = [
        _RESOURCE_0.collect_goal,  # defect
        _RESOURCE_0.interact_goal,  # continue defecting
        _RESOURCE_1.collect_goal,  # cooperate
        _RESOURCE_1.collect_goal,  # cooperate
        _RESOURCE_1.collect_goal,  # continue cooperating
        _RESOURCE_1.interact_goal,  # continue cooperating
        _RESOURCE_0.interact_goal,  # defect
    ]
    actual, _ = _goals_from_observations(puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)

  def test_tremble(self):
    puppeteer = in_the_matrix.Corrigible(
        threshold=1,
        cooperate_resource=_RESOURCE_1,
        defect_resource=_RESOURCE_0,
        margin=2,
        tremble_probability=1,
    )
    inventories = [
        (2, 0, 1),  # not ready to interact
        (4, 1, 0),  # ready to interact if defecting
        (2, 1, 0),  # not ready to interact
        (1, 2, 0),  # not ready to interact
        (2, 1, 0),  # not ready to interact
        (1, 4, 0),  # ready to interact if cooperating
        (3, 1, 0),  # ready to interact if defecting
    ]
    interactions = [
        ((-1, -1, -1), (0, 1, 0)),  # coplayer cooperates
        ((-1, -1, -1), (0, 0, 1)),  # coplayer plays other
        ((-1, -1, -1), (1, 0, 0)),  # coplayer defects
        ((-1, -1, -1), (0, 1, 0)),  # coplayer cooperates
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (1, 1, 0)),  # coplayer cooperates and defects
        ((-1, -1, -1), (2, 0, 1)),  # coplayer defects
    ]
    expected = [
        _RESOURCE_0.collect_goal,  # defect
        _RESOURCE_0.interact_goal,  # continue defecting
        _RESOURCE_0.collect_goal,  # cooperate but tremble and defect
        _RESOURCE_0.collect_goal,  # cooperate but tremble and defect
        _RESOURCE_0.collect_goal,  # continue defecting
        _RESOURCE_0.collect_goal,  # continue defecting
        _RESOURCE_1.collect_goal,  # defect but tremble and cooperate
    ]
    actual, _ = _goals_from_observations(puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)


class RespondToPreviousTest(parameterized.TestCase):

  def test(self):
    puppeteer = in_the_matrix.RespondToPrevious(
        responses={
            _RESOURCE_0: _RESOURCE_2,
            _RESOURCE_1: _RESOURCE_0,
            _RESOURCE_2: _RESOURCE_1,
        },
        margin=1,
    )
    inventories = [
        (1, 1, 1),
        (1, 2, 1),
        (1, 2, 3),
        (2, 3, 1),
        (3, 2, 1),
        (3, 2, 1),
        (2, 3, 1),
    ]
    interactions = [
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (1, 0, 0)),  # opponent plays 0
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (0, 1, 0)),  # opponent plays 1
        ((-1, -1, -1), (1, 1, 1)),  # no clear interaction
        ((-1, -1, -1), (0, 0, 1)),  # opponent plays 2
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
    ]
    expected = [
        mock.ANY,  # random
        _RESOURCE_2.collect_goal,
        _RESOURCE_2.interact_goal,
        _RESOURCE_0.collect_goal,
        _RESOURCE_0.interact_goal,
        _RESOURCE_1.collect_goal,
        _RESOURCE_1.interact_goal,
    ]
    actual, _ = _goals_from_observations(puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)


class AlternatingSpecialistTest(parameterized.TestCase):

  def testOneInteractionPerOption(self):
    puppeteer = in_the_matrix.AlternatingSpecialist(
        targets=[_RESOURCE_0, _RESOURCE_1, _RESOURCE_2],
        interactions_per_target=1,
        margin=1,
    )
    inventories = [
        (1, 1, 1),
        (1, 2, 1),
        (1, 2, 3),
        (2, 3, 1),
        (3, 2, 1),
        (3, 2, 1),
        (2, 3, 1),
    ]
    interactions = [
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (1, 0, 0)),  # opponent plays 0
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (0, 1, 0)),  # opponent plays 1
        ((-1, -1, -1), (2, 2, 2)),  # no clear interaction
        ((-1, -1, -1), (0, 0, 1)),  # opponent plays 2
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
    ]
    expected = [
        _RESOURCE_0.collect_goal,
        _RESOURCE_1.interact_goal,
        _RESOURCE_1.collect_goal,
        _RESOURCE_2.collect_goal,
        _RESOURCE_0.interact_goal,
        _RESOURCE_1.collect_goal,
        _RESOURCE_1.interact_goal,
    ]
    actual, _ = _goals_from_observations(
        puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)

  def testTwoInteractionsPerOption(self):
    puppeteer = in_the_matrix.AlternatingSpecialist(
        targets=[_RESOURCE_0, _RESOURCE_1, _RESOURCE_2],
        interactions_per_target=2,
        margin=1,
    )
    inventories = [
        (1, 1, 1),
        (1, 1, 1),
        (1, 2, 1),
        (1, 2, 1),
        (1, 2, 3),
        (1, 2, 3),
        (2, 3, 1),
        (2, 3, 1),
        (3, 2, 1),
        (3, 2, 1),
        (3, 2, 1),
        (3, 2, 1),
        (2, 3, 1),
        (2, 3, 1),
    ]
    interactions = [
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (1, 0, 0)),  # opponent plays 0
        ((-1, -1, -1), (1, 0, 0)),  # opponent plays 0
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (0, 1, 0)),  # opponent plays 1
        ((-1, -1, -1), (0, 1, 0)),  # opponent plays 1
        ((-1, -1, -1), (2, 2, 2)),  # no clear interaction
        ((-1, -1, -1), (2, 2, 2)),  # no clear interaction
        ((-1, -1, -1), (0, 0, 1)),  # opponent plays 2
        ((-1, -1, -1), (0, 0, 1)),  # opponent plays 2
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
        ((-1, -1, -1), (-1, -1, -1)),  # no interaction
    ]
    expected = [
        _RESOURCE_0.collect_goal,
        _RESOURCE_0.collect_goal,
        _RESOURCE_0.collect_goal,
        _RESOURCE_1.interact_goal,
        _RESOURCE_1.collect_goal,
        _RESOURCE_1.collect_goal,
        _RESOURCE_1.interact_goal,
        _RESOURCE_2.collect_goal,
        _RESOURCE_2.collect_goal,
        _RESOURCE_0.interact_goal,
        _RESOURCE_0.interact_goal,
        _RESOURCE_1.collect_goal,
        _RESOURCE_1.interact_goal,
        _RESOURCE_1.interact_goal,
    ]
    actual, _ = _goals_from_observations(
        puppeteer, inventories, interactions)
    self.assertEqual(actual, expected)


if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common configuration for all territory_* substrates.

See _Territory: Open_ for the general description of the mechanics at play in
this substrate.
"""

from typing import Any, Mapping, Sequence

from ml_collections import config_dict

from meltingpot.utils.substrates import colors
from meltingpot.utils.substrates import game_object_utils
from meltingpot.utils.substrates import shapes
from meltingpot.utils.substrates import specs

PrefabConfig = game_object_utils.PrefabConfig

# Warning: setting `_ENABLE_DEBUG_OBSERVATIONS = True` may cause slowdown.
_ENABLE_DEBUG_OBSERVATIONS = False

_COMPASS = ["N", "E", "S", "W"]

MARKING_SPRITE = """
oxxxxxxo
xoxxxxox
xxoxxoxx
xxxooxxx
xxxooxxx
xxoxxoxx
xoxxxxox
oxxxxxxo
"""


def get_marking_palette(alpha: float) -> Mapping[str, Sequence[int]]:
  alpha_uint8 = int(alpha * 255)
  assert alpha_uint8 >= 0.0 and alpha_uint8 <= 255, "Color value out of range."
  return {"x": shapes.ALPHA, "o": (0, 0, 0, alpha_uint8)}

FLOOR = {
    "name": "floor",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "floor",
                "stateConfigs": [{
                    "state": "floor",
                    "layer": "background",
                    "sprite": "Floor",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Floor",],
                "spriteShapes": [shapes.GRAINY_FLOOR],
                "palettes": [{
                    "*": (27, 22, 20, 255),
                    "+": (23, 17, 15, 255),
                }],
                "noRotates": [True]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

WALL = {
    "name": "wall",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "wall",
                "stateConfigs": [{
                    "state": "wall",
                    "layer": "upperPhysical",
                    "sprite": "Wall",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["Wall",],
                "spriteShapes": [shapes.FILL],
                "palettes": [{"i": (61, 57, 55, 255)}],
                "noRotates": [False]
            }
        },
        {
            "component": "Transform",
        },
        {
            "component": "AllBeamBlocker",
            "kwargs": {}
        },
    ]
}

WALL_HIGHLIGHT_NW = {
    "name": "nw_highlight",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "nw_highlight",
                "stateConfigs": [{
                    "state": "nw_highlight",
                    "layer": "overlay",
                    "sprite": "NWHighlight",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NWHighlight",],
                "spriteShapes": [shapes.NW_HIGHLIGHT],
                "palettes": [shapes.HIGHLIGHT_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

WALL_HIGHLIGHT_E_W = {
    "name": "e_w_highlight",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "e_w_highlight",
                "stateConfigs": [{
                    "state": "e_w_highlight",
                    "layer": "overlay",
                    "sprite": "EWHighlight",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["EWHighlight",],
                "spriteShapes": [shapes.E_W_HIGHLIGHT],
                "palettes": [shapes.HIGHLIGHT_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

WALL_HIGHLIGHT_N_S = {
    "name": "n_s_highlight",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "n_s_highlight",
                "stateConfigs": [{
                    "state": "n_s_highlight",
                    "layer": "overlay",
                    "sprite": "NSHighlight",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NSHighlight",],
                "spriteShapes": [shapes.N_S_HIGHLIGHT],
                "palettes": [shapes.HIGHLIGHT_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

WALL_HIGHLIGHT_NE = {
    "name": "ne_highlight",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "ne_highlight",
                "stateConfigs": [{
                    "state": "ne_highlight",
                    "layer": "overlay",
                    "sprite": "NEHighlight",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["NEHighlight",],
                "spriteShapes": [shapes.NE_HIGHLIGHT],
                "palettes": [shapes.HIGHLIGHT_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

WALL_HIGHLIGHT_SE = {
    "name": "se_highlight",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "se_highlight",
                "stateConfigs": [{
                    "state": "se_highlight",
                    "layer": "overlay",
                    "sprite": "SEHighlight",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SEHighlight",],
                "spriteShapes": [shapes.SE_HIGHLIGHT],
                "palettes": [shapes.HIGHLIGHT_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

WALL_HIGHLIGHT_SW = {
    "name": "sw_highlight",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "sw_highlight",
                "stateConfigs": [{
                    "state": "sw_highlight",
                    "layer": "overlay",
                    "sprite": "SWHighlight",
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "ascii_shape",
                "spriteNames": ["SWHighlight",],
                "spriteShapes": [shapes.SW_HIGHLIGHT],
                "palettes": [shapes.HIGHLIGHT_PALETTE],
                "noRotates": [False]
            }
        },
        {
            "component": "Transform",
        },
    ]
}

SPAWN_POINT = {
    "name": "spawn_point",
    "components": [
        {
            "component": "StateManager",
            "kwargs": {
                "initialState": "playerSpawnPoint",
                "stateConfigs": [{
                    "state": "playerSpawnPoint",
                    "layer": "logic",
                    "groups": ["spawnPoints"],
                }],
            }
        },
        {
            "component": "Appearance",
            "kwargs": {
                "renderMode": "invisible",
                "spriteNames": [],
                "spriteRGBColors": []
            }
        },
        {
            "component": "Transform",
        },
    ]
}


def get_dry_painted_wall_palette(base_color: shapes.Color
                                 ) -> Mapping[str, shapes.ColorRGBA]:
  return {
      "*": shapes.scale_color(base_color, 0.75, 200),
      "#": shapes.scale_color(base_color, 0.90, 150),
  }


def get_brush_palette(
    base_color: shapes.Color) -> Mapping[str, shapes.ColorRGBA]:
  return {
      "*": base_color + (255,),
      "&": shapes.scale_color(base_color, 0.75, 255),
      "o": shapes.scale_color(base_color, 0.55, 255),
      "O": (70, 70, 70, 255),
      "-": (143, 96, 74, 255),
      "+": (117, 79, 61, 255),
      "k": (199, 176, 135, 255),
      "x": shapes.ALPHA,
  }

PLAYER_COLOR_PALETTES = []
BRUSH_PALETTES = []
for human_readable_color in colors.human_readable:
  PLAYER_COLOR_PALETTES.append(shapes.get_palette(human_readable_color))
  BRUSH_PALETTES.append(get_brush_palette(human_readable_color))


def create_resource(num_players: int) -> PrefabConfig:
  """Configure the prefab to use for all resource objects."""
  # Setup unique states corresponding to each player who can claim the resource.
  claim_state_configs = []
  claim_sprite_names = []
  claim_sprite_rgb_colors = []
  for player_idx in range(num_players):
    lua_player_idx = player_idx + 1
    player_color = colors.human_readable[player_idx]
    wet_sprite_name = "Color" + str(lua_player_idx) + "ResourceSprite"
    claim_state_configs.append({
        "state": "claimed_by_" + str(lua_player_idx),
        "layer": "upperPhysical",
        "sprite": wet_sprite_name,
        "groups": ["claimedResources"]
    })
    claim_sprite_names.append(wet_sprite_name)
    # Use alpha channel to make transparent version of claiming agent's color.
    wet_paint_color = player_color + (75,)
    claim_sprite_rgb_colors.append(wet_paint_color)

  prefab = {
      "name": "resource",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "unclaimed",
                  "stateConfigs": [
                      {"state": "unclaimed",
                       "layer": "upperPhysical",
                       "sprite": "UnclaimedResourceSprite"},
                      {"state": "destroyed"},
                  ] + claim_state_configs,
              }
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "colored_square",
                  "spriteNames": claim_sprite_names,
                  "spriteRGBColors": claim_sprite_rgb_colors
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Resource",
              "kwargs": {
                  "initialHealth": 2,
                  "destroyedState": "destroyed",
                  "reward": 1.0,
                  "rewardRate": 0.01,
                  "rewardDelay": 25,
                  "delayTillSelfRepair": 15,
                  "selfRepairProbability": 0.1,
              }
          },
      ]
  }
  return prefab


def create_resource_texture() -> PrefabConfig:
  """Configure the background texture for a resource. It looks like a wall."""
  prefab = {
      "name": "resource_texture",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "unclaimed",
                  "stateConfigs": [
                      {"state": "unclaimed",
                       "layer": "lowerPhysical",
                       "sprite": "UnclaimedResourceSprite"},
                      {"state": "destroyed"},
                  ],
              }
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["UnclaimedResourceSprite",],
                  "spriteShapes": [shapes.WALL],
                  "palettes": [{"*": (61, 61, 61, 255),
                                "#": (80, 80, 80, 255)}],
                  "noRotates": [True]
              }
          },
          {
              "component": "Transform",
          },
      ]
  }
  return prefab


def create_reward_indicator(num_players) -> PrefabConfig:
  """Configure object indicating if a resource is currently providing reward."""
  # Setup unique states corresponding to each player who can claim the resource.
  claim_state_configs = []
  claim_sprite_names = []
  claim_sprite_shapes = []
  claim_palettes = []
  claim_no_rotates = []
  for player_idx in range(num_players):
    lua_player_idx = player_idx + 1
    player_color = colors.human_readable[player_idx]
    dry_sprite_name = "Color" + str(lua_player_idx) + "DryPaintSprite"
    claim_state_configs.append({
        "state": "dry_claimed_by_" + str(lua_player_idx),
        "layer": "overlay",
        "sprite": dry_sprite_name,
    })
    claim_sprite_names.append(dry_sprite_name)
    claim_sprite_shapes.append(shapes.WALL)
    claim_palettes.append(get_dry_painted_wall_palette(player_color))
    claim_no_rotates.append(True)
  prefab = {
      "name": "reward_indicator",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "inactive",
                  "stateConfigs": [
                      {"state": "inactive"},
                  ] + claim_state_configs,
              }
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": claim_sprite_names,
                  "spriteShapes": claim_sprite_shapes,
                  "palettes": claim_palettes,
                  "noRotates": claim_no_rotates
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "RewardIndicator",
          },
      ]
  }
  return prefab


def create_damage_indicator() -> PrefabConfig:
  """Configure the object indicating whether or not a resource is damaged."""
  damaged_resource_sprite = """
      ,,bb,,,,
      ,,bb,bb,
      ,,,b,,b,
      ,,,b,,,,
      ,,,b,,,b
      ,,,bb,,b
      ,,,bb,,b
      b,,,b,,,
  """
  damaged_resource_palette = {",": shapes.ALPHA, "b": shapes.BLACK}
  prefab = {
      "name": "damage_indicator",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "inactive",
                  "stateConfigs": [
                      {"state": "inactive",
                       "layer": "superDirectionIndicatorLayer"},
                      {"state": "damaged",
                       "layer": "superDirectionIndicatorLayer",
                       "sprite": "DamagedResource"},
                  ],
              }
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["DamagedResource"],
                  "spriteShapes": [damaged_resource_sprite],
                  "palettes": [damaged_resource_palette],
                  "noRotates": [True]
              }
          },
          {
              "component": "Transform",
          },
      ]
  }
  return prefab


def create_prefabs(num_players: int):
  """Returns the prefabs dictionary."""
  prefabs = {
      "spawn_point": SPAWN_POINT,
      "floor": FLOOR,
      "wall": WALL,
      "wall_highlight_nw": WALL_HIGHLIGHT_NW,
      "wall_highlight_e_w": WALL_HIGHLIGHT_E_W,
      "wall_highlight_n_s": WALL_HIGHLIGHT_N_S,
      "wall_highlight_ne": WALL_HIGHLIGHT_NE,
      "wall_highlight_se": WALL_HIGHLIGHT_SE,
      "wall_highlight_sw": WALL_HIGHLIGHT_SW,
      "resource": create_resource(num_players=num_players),
      "resource_texture": create_resource_texture(),
      "reward_indicator": create_reward_indicator(num_players),
      "damage_indicator": create_damage_indicator(),
  }
  return prefabs


# Primitive action components.
# pylint: disable=bad-whitespace
# pyformat: disable
NOOP       = {"move": 0, "turn":  0, "fireZap": 0, "fireClaim": 0}
FORWARD    = {"move": 1, "turn":  0, "fireZap": 0, "fireClaim": 0}
STEP_RIGHT = {"move": 2, "turn":  0, "fireZap": 0, "fireClaim": 0}
BACKWARD   = {"move": 3, "turn":  0, "fireZap": 0, "fireClaim": 0}
STEP_LEFT  = {"move": 4, "turn":  0, "fireZap": 0, "fireClaim": 0}
TURN_LEFT  = {"move": 0, "turn": -1, "fireZap": 0, "fireClaim": 0}
TURN_RIGHT = {"move": 0, "turn":  1, "fireZap": 0, "fireClaim": 0}
FIRE_ZAP   = {"move": 0, "turn":  0, "fireZap": 1, "fireClaim": 0}
FIRE_CLAIM = {"move": 0, "turn":  0, "fireZap": 0, "fireClaim": 1}
# pyformat: enable
# pylint: enable=bad-whitespace

ACTION_SET = (
    NOOP,
    FORWARD,
    BACKWARD,
    STEP_LEFT,
    STEP_RIGHT,
    TURN_LEFT,
    TURN_RIGHT,
    FIRE_ZAP,
    FIRE_CLAIM
)


# The Scene object is a non-physical object, its components implement global
# logic.
def create_scene():
  """Creates the global scene."""
  scene = {
      "name": "scene",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "scene",
                  "stateConfigs": [{
                      "state": "scene",
                  }],
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "StochasticIntervalEpisodeEnding",
              "kwargs": {
                  "minimumFramesPerEpisode": 1000,
                  "intervalLength": 100,  # Set equal to unroll length.
                  "probabilityTerminationPerInterval": 0.2
              }
          }
      ]
  }
  return scene


def create_avatar_object(player_idx: int) -> Mapping[str, Any]:
  """Create an avatar object that always sees itself as blue."""
  # Lua is 1-indexed.
  lua_index = player_idx + 1

  color_palette = PLAYER_COLOR_PALETTES[player_idx]
  paintbrush_palette = BRUSH_PALETTES[player_idx]
  live_state_name = "player{}".format(lua_index)
  avatar_sprite_name = "avatarSprite{}".format(lua_index)
  avatar_object = {
      "name": "avatar",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": live_state_name,
                  "stateConfigs": [
                      # Initial player state.
                      {"state": live_state_name,
                       "layer": "upperPhysical",
                       "sprite": avatar_sprite_name,
                       "contact": "avatar",
                       "groups": ["players"]},

                      # Player wait state used when they have been zapped out.
                      {"state": "playerWait",
                       "groups": ["playerWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": [avatar_sprite_name],
                  "spriteShapes": [shapes.CUTE_AVATAR_HOLDING_PAINTBRUSH],
                  "palettes": [{**color_palette, **paintbrush_palette}],
                  "noRotates": [True]
              }
          },
          {
              "component": "Avatar",
              "kwargs": {
                  "index": lua_index,
                  "aliveState": live_state_name,
                  "waitState": "playerWait",
                  "spawnGroup": "spawnPoints",
                  "actionOrder": ["move",
                                  "turn",
                                  "fireZap",
                                  "fireClaim"],
                  "actionSpec": {
                      "move": {"default": 0, "min": 0, "max": len(_COMPASS)},
                      "turn": {"default": 0, "min": -1, "max": 1},
                      "fireZap": {"default": 0, "min": 0, "max": 1},
                      "fireClaim": {"default": 0, "min": 0, "max": 1},
                  },
                  "view": {
                      "left": 5,
                      "right": 5,
                      "forward": 9,
                      "backward": 1,
                      "centered": False
                  },
              }
          },
          {
              "component": "Paintbrush",
              "kwargs": {
                  "shape": shapes.PAINTBRUSH,
                  "palette": paintbrush_palette,
                  "playerIndex": lua_index,
              }
          },
          {
              "component": "Zapper",
              "kwargs": {
                  "cooldownTime": 4,
                  "beamLength": 2,
                  "beamRadius": 1,
                  "framesTillRespawn": 1e6,  # Effectively never respawn.
                  "penaltyForBeingZapped": 0,
                  "rewardForZapping": 0,
                  # GraduatedSanctionsMarking handles removal instead of Zapper.
                  "removeHitPlayer": False,
              }
          },
          {
              "component": "ReadyToShootObservation",
          },
          {
              "component": "ResourceClaimer",
              "kwargs": {
                  "color": color_palette["*"],
                  "playerIndex": lua_index,
                  "beamLength": 2,
                  "beamRadius": 0,
                  "beamWait": 0,
              }
          },
          {
              "component": "Taste",
              "kwargs": {
                  "role": "none",
                  "rewardAmount": 1.0,
              }
          },
      ]
  }
  if _ENABLE_DEBUG_OBSERVATIONS:
    avatar_object["components"].append({
        "component": "LocationObserver",
        "kwargs": {"objectIsAvatar": True, "alsoReportOrientation": True},
    })

  return avatar_object


def create_marking_overlay(player_idx: int) -> Mapping[str, Any]:
  """Create a graduated sanctions marking overlay object."""
  # Lua is 1-indexed.
  lua_idx = player_idx + 1

  marking_object = {
      "name": "avatar_marking",
      "components": [
          {
              "component": "StateManager",
              "kwargs": {
                  "initialState": "avatarMarkingWait",
                  "stateConfigs": [
                      # Declare one state per level of the hit logic.
                      {"state": "level_1",
                       "layer": "superOverlay",
                       "sprite": "sprite_for_level_1"},
                      {"state": "level_2",
                       "layer": "superOverlay",
                       "sprite": "sprite_for_level_2"},

                      # Invisible inactive (zapped out) overlay type.
                      {"state": "avatarMarkingWait",
                       "groups": ["avatarMarkingWaits"]},
                  ]
              }
          },
          {
              "component": "Transform",
          },
          {
              "component": "Appearance",
              "kwargs": {
                  "renderMode": "ascii_shape",
                  "spriteNames": ["sprite_for_level_1",
                                  "sprite_for_level_2"],
                  "spriteShapes": [MARKING_SPRITE,
                                   MARKING_SPRITE],
                  "palettes": [get_marking_palette(0.0),
                               get_marking_palette(1.0)],
                  "noRotates": [True] * 3
              }
          },
          {
              "component": "GraduatedSanctionsMarking",
              "kwargs": {
                  "playerIndex": lua_idx,
                  "waitState": "avatarMarkingWait",
                  "hitName": "zapHit",
                  "recoveryTime": 50,
                  "hitLogic": [
                      {"levelIncrement": 1,
                       "sourceReward": 0,
                       "targetReward": 0,
                       "freeze": 25},
                      {"levelIncrement": -1,
                       "sourceReward": 0,
                       "targetReward": 0,
                       "remove": True}
                  ],
              }
          },
      ]
  }
  return marking_object


def create_avatar_and_associated_objects(num_players):
  """Returns list of avatars and their associated marking objects."""
  avatar_objects = []
  additional_objects = []
  for player_idx in range(0, num_players):
    game_object = create_avatar_object(player_idx)
    avatar_objects.append(game_object)

    marking_object = create_marking_overlay(player_idx)
    additional_objects.append(marking_object)

  return avatar_objects + additional_objects


def get_config():
  """Default configuration."""
  config = config_dict.ConfigDict()

  # Action set configuration.
  config.action_set = ACTION_SET
  # Observation format configuration.
  config.individual_observation_names = [
      "RGB",
      "READY_TO_SHOOT",
  ]
  config.global_observation_names = [
      "WORLD.RGB",
  ]

  # The specs of the environment (from a single-agent perspective).
  config.action_spec = specs.action(len(ACTION_SET))
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "READY_TO_SHOOT": specs.OBSERVATION["READY_TO_SHOOT"],
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(120, 184),
  })

  config.valid_roles = frozenset({"default"})

  return config


def build(
    roles: Sequence[str],
    config: config_dict.ConfigDict,
) -> Mapping[str, Any]:
  """Build substrate definition given player roles."""
  num_players = len(roles)
  # Build the rest of the substrate definition.
  substrate_definition = dict(
      levelName="territory",
      levelDirectory="meltingpot/lua/levels",
      numPlayers=num_players,
      # Define upper bound of episode length since episodes end stochastically.
      maxEpisodeLengthFrames=5000,
      spriteSize=8,
      topology=config.layout.topology,  # Choose from ["BOUNDED", "TORUS"],
      simulation={
          "map": config.layout.ascii_map,
          "gameObjects": create_avatar_and_associated_objects(num_players),
          "scene": create_scene(),
          "prefabs": create_prefabs(num_players),
          "charPrefabMap": config.layout.char_prefab_map,
      },
  )
  return substrate_definition



# Copyright 2020 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for multiplayer_wrapper."""

from unittest import mock

from absl.testing import absltest
import dm_env
import dmlab2d
import numpy as np

from meltingpot.utils.substrates.wrappers import collective_reward_wrapper

RGB_SPEC = dm_env.specs.Array(shape=(2, 1), dtype=np.int8)
COLLECTIVE_REWARD_SPEC = dm_env.specs.Array(shape=(), dtype=np.float64)

NUM_PLAYERS = 3
REWARDS = [1.0, 2.0, 3.0]
RGB = np.zeros((2, 1))


class CollectiveRewardWrapperTest(absltest.TestCase):

  def test_get_timestep(self):
    env = mock.Mock(spec_set=dmlab2d.Environment)
    wrapped = collective_reward_wrapper.CollectiveRewardWrapper(env)

    source = dm_env.TimeStep(
        step_type=dm_env.StepType.MID,
        reward=REWARDS,
        discount=1.0,
        observation=[{'RGB': RGB} for _ in range(NUM_PLAYERS)])
    actual = wrapped._get_timestep(source)
    added_key = collective_reward_wrapper._COLLECTIVE_REWARD_OBS
    collective_reward = np.sum(REWARDS)
    expected_observation = [
        {'RGB': RGB, added_key: collective_reward},
    ] * NUM_PLAYERS
    expected_timestep = dm_env.TimeStep(
        step_type=dm_env.StepType.MID,
        reward=REWARDS,
        discount=1.0,
        observation=expected_observation)
    np.testing.assert_equal(actual, expected_timestep)

  def test_spec(self):
    env = mock.Mock(spec_set=dmlab2d.Environment)
    env.observation_spec.return_value = [{
        'RGB': RGB_SPEC,
    }] * NUM_PLAYERS
    wrapped = collective_reward_wrapper.CollectiveRewardWrapper(env)

    added_key = collective_reward_wrapper._COLLECTIVE_REWARD_OBS
    self.assertEqual(wrapped.observation_spec(), [
        {'RGB': RGB_SPEC, added_key: COLLECTIVE_REWARD_SPEC}] * NUM_PLAYERS)

if __name__ == '__main__':
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration for predator_prey__open.

Example video: https://youtu.be/0ZlrkWsWzMw

See predator_prey.py for a detailed description applicable to all predator_prey
substrates.

In this variant prey must forage over a large field of apples and acorns in the
center of the map. Since the space is so open it should be possible for the prey
to move together in larger groups so they can defend themselves from predators.
Another prey strategy focused on acorns instead of apples is also possible. In
this case prey collect acorns and bring them back to safe tall grass to consume
them.
"""

from ml_collections import config_dict

from meltingpot.configs.substrates import predator_prey as base_config
from meltingpot.utils.substrates import specs

build = base_config.build

ASCII_MAP = """
/;___________________,/
;]*******************[,
!vvvvvvvvvvvvvvvvvvvvv|
!'''''''''''''''''''''|
!''XXXXXXXXXXXXXXXXX''|
!''XAaaaaaaaaaaAaaaX''|
!''Xaaaa&aaaAaaaaaaX''|
!'aaaaaaaaaaaaaaaaaaa'|
!Aaaaaaaaaaaaaaaaaaaaa|
!aaaaaaaaaaaaaaAaaaaaa|
!aAaaaaaaaaaaaaaaa&aaA|
!'aaaaaaAaaaaaaaaaAaa'|
!''Xaaaaaaa&aaaaaaaX''|
!''XaaaaaaaaAaaaaaaX''|
!''XXXXXXXXXXXXXXXXX''|
!'''''''''''''''''''''|
!^^^^^^^^^^^^^^^^^^^^^|
L+*******************=J
/L~~~~~~~~~~~~~~~~~~~J/
"""

# `prefab` determines which prefab game object to use for each `char` in the
# ascii map.
CHAR_PREFAB_MAP = {
    "*": {"type": "all", "list": ["safe_grass", "spawn_point_prey"]},
    "&": {"type": "all", "list": ["tiled_floor", "apple", "spawn_point_prey"]},
    "X": {"type": "all", "list": ["tiled_floor", "spawn_point_predator"]},
    "a": {"type": "all", "list": ["tiled_floor", "apple"]},
    "A": {"type": "all", "list": ["tiled_floor", "floor_acorn"]},
    ";": "nw_wall_corner",
    ",": "ne_wall_corner",
    "J": "se_wall_corner",
    "L": "sw_wall_corner",
    "_": "wall_north",
    "|": "wall_east",
    "~": "wall_south",
    "!": "wall_west",
    "=": "nw_inner_wall_corner",
    "+": "ne_inner_wall_corner",
    "]": "se_inner_wall_corner",
    "[": "sw_inner_wall_corner",
    "'": "tiled_floor",
    "#": "safe_grass",
    "<": "safe_grass_w_edge",
    "^": "safe_grass_n_edge",
    ">": "safe_grass_e_edge",
    "v": "safe_grass_s_edge",
    "l": "safe_grass_ne_corner",
    "j": "safe_grass_se_corner",
    "z": "safe_grass_sw_corner",
    "r": "safe_grass_nw_corner",
    "/": "fill",
}


def get_config():
  """Default configuration."""
  config = base_config.get_config()

  # Override the map layout settings.
  config.layout = config_dict.ConfigDict()
  config.layout.ascii_map = ASCII_MAP
  config.layout.char_prefab_map = CHAR_PREFAB_MAP

  # The specs of the environment (from a single-agent perspective).
  config.timestep_spec = specs.timestep({
      "RGB": specs.OBSERVATION["RGB"],
      "STAMINA": specs.float64(),
      # Debug only (do not use the following observations in policies).
      "WORLD.RGB": specs.rgb(152, 184),
  })

  # The roles assigned to each player.
  config.default_player_roles = ("predator",) * 3 + ("prey",) * 10

  return config



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helpers for defining substrate specs.

Used to allow substrates to easily define the single-player specs within their
configs.
"""

from typing import Mapping, Optional

import dm_env
import immutabledict
import numpy as np

STEP_TYPE = dm_env.specs.BoundedArray(
    shape=(),
    dtype=np.int64,
    minimum=min(dm_env.StepType),
    maximum=max(dm_env.StepType),
    name='step_type',
)
DISCOUNT = dm_env.specs.BoundedArray(
    shape=(), dtype=np.float64, minimum=0, maximum=1, name='discount')
REWARD = dm_env.specs.Array(shape=(), dtype=np.float64, name='reward')
OBSERVATION = immutabledict.immutabledict({
    'READY_TO_SHOOT': dm_env.specs.Array(
        shape=(), dtype=np.float64, name='READY_TO_SHOOT'),
    'RGB': dm_env.specs.Array(shape=(88, 88, 3), dtype=np.uint8, name='RGB'),
    'POSITION': dm_env.specs.Array(shape=(2,), dtype=np.int32, name='POSITION'),
    'ORIENTATION': dm_env.specs.Array(
        shape=(), dtype=np.int32, name='ORIENTATION'),
})
_ACTION = dm_env.specs.DiscreteArray(
    num_values=1, dtype=np.int64, name='action')


def float32(*shape: int, name: Optional[str] = None) -> dm_env.specs.Array:
  """Returns the spec for an np.float32 tensor.

  Args:
    *shape: the shape of the tensor.
    name: optional name for the spec.
  """
  return dm_env.specs.Array(shape=shape, dtype=np.float32, name=name)


def float64(*shape: int, name: Optional[str] = None) -> dm_env.specs.Array:
  """Returns the spec for an np.float64 tensor.

  Args:
    *shape: the shape of the tensor.
    name: optional name for the spec.
  """
  return dm_env.specs.Array(shape=shape, dtype=np.float64, name=name)


def int32(*shape: int, name: Optional[str] = None) -> dm_env.specs.Array:
  """Returns the spec for an np.int32 tensor.

  Args:
    *shape: the shape of the tensor.
    name: optional name for the spec.
  """
  return dm_env.specs.Array(shape=shape, dtype=np.int32, name=name)


def int64(*shape: int, name: Optional[str] = None) -> dm_env.specs.Array:
  """Returns the spec for an np.int32 tensor.

  Args:
    *shape: the shape of the tensor.
    name: optional name for the spec.
  """
  return dm_env.specs.Array(shape=shape, dtype=np.int64, name=name)


def action(num_actions: int) -> dm_env.specs.DiscreteArray:
  """Returns the spec for an action.

  Args:
    num_actions: the number of actions that can be taken.
  """
  return _ACTION.replace(num_values=num_actions)


def rgb(height: int,
        width: int,
        name: Optional[str] = 'RGB') -> dm_env.specs.Array:
  """Returns the spec for an RGB observation.

  Args:
    height: the height of the observation.
    width: the width of the observation.
    name: optional name for the spec.
  """
  return OBSERVATION['RGB'].replace(shape=(height, width, 3), name=name)


def world_rgb(ascii_map: str,
              sprite_size: int,
              name: Optional[str] = 'WORLD.RGB') -> dm_env.specs.Array:
  """Returns the spec for a WORLD.RGB observation.

  Args:
    ascii_map: the height of the observation.
    sprite_size: the width of the observation.
    name: optional name for the spec.
  """
  lines = ascii_map.strip().split('\n')
  height = len(lines) * sprite_size
  width = len(lines[0]) * sprite_size if height else 0
  return rgb(height, width, name)


def inventory(num_resources: int,
              name: Optional[str] = 'INVENTORY') -> dm_env.specs.Array:
  """Returns the spec for an INVENTORY observation.

  Args:
    num_resources: the number of resource types in the inventory.
    name: optional name for the spec.
  """
  return float64(num_resources, name=name)


def interaction_inventories(
    num_resources: int,
    name: Optional[str] = 'INTERACTION_INVENTORIES') -> dm_env.specs.Array:
  """Returns the spec for an INTERACTION_INVENTORIES observation.

  Args:
    num_resources: the number of resource types in the inventory.
    name: optional name for the spec.
  """
  return float64(2, num_resources, name=name)


def timestep(
    observation_spec: Mapping[str, dm_env.specs.Array]) -> dm_env.TimeStep:
  """Returns the spec for a timestep.

  Args:
    observation_spec: the observation spec. Spec names will be overwritten with
      their key.
  """
  observation_spec = immutabledict.immutabledict({
      name: spec.replace(name=name) for name, spec in observation_spec.items()
  })
  return dm_env.TimeStep(
      step_type=STEP_TYPE,
      discount=DISCOUNT,
      reward=REWARD,
      observation=observation_spec,
  )



# Copyright 2022 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A human player for testing fruit_market.

Note: The real agents can make and accept offers up to size 3 (up to 3 apples
for up to 3 bananas). However this human player script only allows offers up to
size 1. The reason is just that we started to run out of keys on the keyboard to
represent higher offers.

Use `WASD` keys to move the player around.
Use `Q and E` to turn the player.
Use `TAB` to switch which player you are controlling.
Use 'Z' to eat an apple from your inventory.
Use 'X' to eat a banana from your inventory.
"""
import argparse
import json
from ml_collections import config_dict
import pygame

from meltingpot.configs.substrates import fruit_market__concentric_rivers
from meltingpot.human_players import level_playing_utils


def get_offer_apple_pressed() -> int:
  """Sets apple offer to either -1, 0, or 1."""
  key_pressed = pygame.key.get_pressed()
  if key_pressed[pygame.K_1]:
    return -1
  if key_pressed[pygame.K_2]:
    return 1
  return 0


def get_offer_banana_pressed() -> int:
  """Sets banana offer to either -1, 0, or 1."""
  key_pressed = pygame.key.get_pressed()
  if key_pressed[pygame.K_3]:
    return -1
  if key_pressed[pygame.K_4]:
    return 1
  return 0


def get_push_pull() -> int:
  """Sets shove to either -1, 0, or 1."""
  if level_playing_utils.get_right_shift_pressed():
    return 1
  if level_playing_utils.get_left_control_pressed():
    return -1
  return 0

environment_configs = {
    'fruit_market__concentric_rivers': fruit_market__concentric_rivers,
}

_ACTION_MAP = {
    # Basic movement actions
    'move': level_playing_utils.get_direction_pressed,
    'turn': level_playing_utils.get_turn_pressed,
    # Trade actions
    'eat_apple': level_playing_utils.get_key_z_pressed,
    'eat_banana': level_playing_utils.get_key_x_pressed,
    'offer_apple': get_offer_apple_pressed,  # 1 and 2
    'offer_banana': get_offer_banana_pressed,  # 3 and 4
    'offer_cancel': level_playing_utils.get_key_number_five_pressed,
    # Grappling actions
    'hold': level_playing_utils.get_space_key_pressed,
    'shove': get_push_pull,
}


def verbose_fn(env_timestep, player_index, current_player_index):
  """Print using this function once enabling the option --verbose=True."""
  lua_index = player_index + 1
  inventory = env_timestep.observation[f'{lua_index}.INVENTORY']
  hunger = env_timestep.observation[f'{lua_index}.HUNGER']
  my_offer = env_timestep.observation[f'{lua_index}.MY_OFFER']
  offers = env_timestep.observation[f'{lua_index}.OFFERS']
  # Only print offer observations from player 0.
  if player_index == current_player_index:
    print(
        f'player: {player_index} --- inventory: {inventory}, hunger: {hunger}')
    print(f'**player 0 view of offers:\n{offers}')
    print(f'**player 0 view of own offer: {my_offer}')


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument(
      '--level_name',
      type=str,
      default='fruit_market__concentric_rivers',
      choices=environment_configs.keys(),
      help='Level name to load')
  parser.add_argument(
      '--observation', type=str, default='RGB', help='Observation to render')
  parser.add_argument(
      '--settings', type=json.loads, default={}, help='Settings as JSON string')
  # Activate verbose mode with --verbose=True.
  parser.add_argument(
      '--verbose', type=bool, default=False, help='Print debug information')
  # Activate events printing mode with --print_events=True.
  parser.add_argument(
      '--print_events', type=bool, default=False, help='Print events')

  args = parser.parse_args()
  env_module = environment_configs[args.level_name]
  env_config = env_module.get_config()
  with config_dict.ConfigDict(env_config).unlocked() as env_config:
    roles = env_config.default_player_roles
    env_config.lab2d_settings = env_module.build(roles, env_config)
  level_playing_utils.run_episode(
      args.observation, args.settings, _ACTION_MAP,
      env_config, level_playing_utils.RenderType.PYGAME,
      verbose_fn=verbose_fn if args.verbose else None,
      print_events=args.print_events)

if __name__ == '__main__':
  main()
